'*****************************************************************************
'*  Name    : TftGraphic                                                     *
'*  Author  : David John Barker                                              *
'*            Swordfish porting by Coccoliso                                 *
'*  Notice  : Copyright (c) 2012 Mecanique                                   *
'*          : All Rights Reserved                                            *
'*  Date    : 7/07/2013                                                      *
'*  Version : 1.1                                                            *
'*  Notes   : 1.1 Support for 8, 16 and 32 bit MCU                           *
'*          : 1.0 Initial release                                            *
'*****************************************************************************
MODULE TftGraphic
  
' orientation...
PUBLIC CONST
   orPortrait   = 0,
   orLandscape  = 1,
   
' pen styles...
   psNormal = 0,
   psDot    = 1,
   psDash   = 2,
   
' brush styles..
   bsClear = 0,
   bsSolid = 1,  
   bsTransp= 2,

' text out alignment constants
   taLeft   = $01,
   taCenter = $02,
   taRight  = $04  

// x, y type
#ifdef GLCD_XY_16
PUBLIC TYPE TXY = WORD
#else
PUBLIC TYPE TXY = BYTE
#endif
   
' color type
PUBLIC STRUCTURE GraphicColor
   Value AS LONGWORD
   Red AS Value.Byte2
   Green AS Value.Byte1
   Blue AS Value.Byte0
END STRUCTURE

' x, y position structure...   
PUBLIC STRUCTURE GraphicPosition
   x AS WORD
   y AS WORD
END STRUCTURE

' Pen structure...
PUBLIC STRUCTURE GraphicPen
   Color AS GraphicColor
   Style AS BYTE
END STRUCTURE  

' Brush structure...
PUBLIC STRUCTURE GraphicBrush
   Color AS GraphicColor
   Style AS BYTE  
END STRUCTURE  

' font structure...
PUBLIC STRUCTURE GraphicFont
   Header AS WORD
   Data AS LONGWORD
   Width AS BYTE
   Height AS BYTE
   LetterSpacing AS BYTE   
   NumberOfChars AS Header.Byte1
   IsFixed AS Header.Booleans(0)
   IsVariable AS Header.Booleans(1)
   IsScanX AS Header.Booleans(6)
   IsScanY AS Header.Booleans(7)
   IsExtendedHeader AS Header.Booleans(3)
END STRUCTURE

' image structure
PUBLIC STRUCTURE GraphicImage
   Header AS BYTE
   Width AS BYTE
   Height AS BYTE
   NumberOfFrames AS BYTE
   IsTransparent AS Header.Booleans(7)
   IsRLE AS Header.Booleans(6)
   Is12Bit AS Header.Booleans(2)
   Is8Bit AS Header.Booleans(1)
   Is1Bit AS Header.Booleans(0)
END STRUCTURE   

' public variables...
PUBLIC DIM 
    Pen AS GraphicPen,
    Brush AS GraphicBrush,
    Font AS GraphicFont, 
    Align AS BYTE
   
' local variables...
PRIVATE DIM 
    _fontTableOffset AS BYTE,
    _fontTableWidth AS BYTE,     
    _charAddress AS WORD

'****************************************************************************
'* Name    : ReadByte                                                       *
'* Purpose : Read a single byte from the font table                         *
'****************************************************************************
PRIVATE FUNCTION ReadByte(index AS WORD) AS BYTE
   TABLEPTR = Font.Data + index
   ASM
   TBLRD *+
   movff TABLAT, ReadByte  
   END ASM 
END FUNCTION  

'****************************************************************************
'* Name    : ReadUShort                                                     *
'* Purpose : Read a UShort from the font table                              *
'****************************************************************************
PRIVATE FUNCTION ReadUShort(index AS WORD) AS WORD
   TABLEPTR = Font.Data + index * 2 
   ASM
   TBLRD *+
   movff TABLAT, ReadUShort
   TBLRD *+
   movff TABLAT, ReadUShort + 1
   END ASM   
END FUNCTION 

'****************************************************************************
'* Name    : LoadCharAddress                                                *
'* Purpose : Load the font table starting address for a given char          *
'****************************************************************************
PUBLIC SUB LoadCharAddress(Value AS BYTE)
   Value = Value - _fontTableOffset  
   IF Font.IsFixed THEN
      _charAddress = Font.Data + Value * _fontTableWidth
   ELSE   
      _charAddress = Font.Data + ReadUShort(Value)
   END IF 
END SUB

'****************************************************************************
'* Name    : ReadFontByte                                                   *
'* Purpose : Read a character byte from the font table. Typically called    *
'*         : following a call to LoadCharAddress()                          *
'****************************************************************************
PUBLIC FUNCTION ReadFontByte() AS BYTE 
   TABLEPTR = _charAddress
   _charAddress = _charAddress + 1
   ASM
   TBLRD *+
   movff TABLAT, ReadFontByte
   END ASM
END FUNCTION  

'****************************************************************************
'* Name    : SetFont                                                        *
'* Purpose : Set the font information block                                 *
'*         : BYTE(0) &H81 - Fixed width font table, scan along Y            *
'*         :         &H82 - Variable width font table, scan along y         *
'*         :         &H41 - Fixed Width Font table, scan along x            *
'*         :         &H42 - Variable Width Font table, scan along x         *
'*         :         NOTE                                                   *
'*         :         If bit 3 is set, then the font table has an extended   *
'*         :         header - see extended header information below         *
'*         : BYTE(1) Number of characters in the font table                 *
'*         : BYTE(2) Character width (0 for variable width fonts)           *
'*         : BYTE(3) Character height                                       *
'*         :                                                                *
'*         : BYTE(4) EXTENDED HEADER                                        *
'*         :         If bit 3 of the main header is set, then byte(4) holds *
'*         :         the size of the extended header block                  *
'*         :                                                                *
'*         : A fixed width font has its character data starting at address  *
'*         : Font.Data. Variable width fonts have a NumberOfChars x UShort  *
'*         : character lookup table starting at address Font.Data           *
'*         : Font tables are assumed to start at character 32 (space). For  *
'*         : font tables that start at a different character, you can       *
'*         : pass an optional TableOffset. For example, a font table that   *
'*         : just contains numbers would typically start at 48 (&H30)       *
'****************************************************************************
PUBLIC SUB SetFont(BYREFCONST GraphicFont() AS BYTE)
   DIM FontHeaderSize AS BYTE

'   Font.Style = FontStyle.IsNormal   ' default font style
   Font.Data = @GraphicFont           ' address of font table
   Font.Header.Byte0 = ReadByte(0)    ' font type, 
   Font.Header.Byte1 = ReadByte(1)    ' number of characters   
   Font.Width = ReadByte(2)           ' character width
   Font.Height = ReadByte(3)          ' character height
   FontHeaderSize = 4                 ' default header size
   _fontTableOffset = 32              ' default character offset
   
   ' has the font table an extended header...
   IF Font.IsExtendedHeader THEN
      FontHeaderSize = FontHeaderSize + ReadByte(4)
      _fontTableOffset = ReadByte(5)
   END IF    

   ' calculate fixed font table width...
   IF Font.IsFixed THEN
      _fontTableWidth = BYTE(Font.Height + 7) / 8 * Font.Width
   END IF
   
   ' skip over header information and point to font table...
   Font.Data = Font.Data + FontHeaderSize   
END SUB

'****************************************************************************
'* Name    : LoadLetterSpacing                                              *
'* Purpose :                                                                *
'****************************************************************************
PUBLIC SUB LoadLetterSpacing()
   IF Font.IsVariable THEN
      _charAddress = Font.Data + ReadUShort(Font.NumberOfChars - 1)
      Font.Width = ReadFontByte
   END IF
END SUB

'****************************************************************************
'* Name    : WidthFontChar                                                  *
'* Purpose : Finds width of a font character                                *
'****************************************************************************
SUB WidthFontChar(fontChar AS CHAR)
   IF Font.IsVariable THEN
      LoadCharAddress(fontChar)
      Font.Width = ReadFontByte
   END IF  
END SUB

{
****************************************************************************
* Name    : TextWidth                                                      *
* Purpose : Find width of a string in pixels                               *
****************************************************************************
} 
PUBLIC FUNCTION TextWidth(pStr AS STRING) AS TXY
   Result = 0
   FSR0 = @pStr
   WHILE INDF0 <> 0
      WidthFontChar(POSTINC0) 
      INC(Result, Font.Width)
      IF Font.IsVariable AND (INDF0 <> 0) THEN
         INC(Result, Font.LetterSpacing)
      ENDIF
   WEND
   IF Font.IsVariable THEN
      DEC(Result, Font.LetterSpacing)
   ENDIF
END FUNCTION

'****************************************************************************
'* Name    : Initialize                                                     *
'* Purpose : Configure the graphics module before use                       *
'****************************************************************************
PRIVATE SUB Initialize()  
   Pen.Color.Value = $FFFFFF
   Brush.Color.Value = 0 
   Font.Header = 0
   Font.LetterSpacing = 1
   Pen.Style = psNormal
   Brush.Style = bsSolid  
   Align = taLeft    
END SUB

Initialize()
           
